from typing import List, Optional, Tuple, Union

from runllm.constants.enums import (
    ImplementationType,
    OpenAIGPTModelType,
    PrimitiveType,
    ResourceType,
    Service,
)
from runllm.context import ContextKey
from runllm.exceptions import (
    InvalidConfigurationException,
    InvalidResourceParamsException,
    UnknownImplementationTypeException,
)
from runllm.primitives.custom_primitive import CustomPrimitive
from runllm.primitives.primitive import Primitive, initialize_primitive_within_task
from runllm.primitives.spec import CustomSpec, DefaultGenerateSpec, FunctionSpec
from runllm.resources.parameters import GenerateParams
from runllm.resources.resources import Resource


class GeneratePrimitive(Primitive):
    inputs: List[Union[str, ContextKey]]
    implementation_spec: Union[DefaultGenerateSpec, CustomSpec]

    def __init__(
        self,
        data: Union[CustomPrimitive, ContextKey],
        implementation_type: ImplementationType,
        resource: Optional[Resource] = None,
        params: Optional[GenerateParams] = None,
        function_spec: Optional[FunctionSpec] = None,
    ):
        dependencies, context_keys, inputs = self._process_data(data)
        spec = self._construct_primitive_spec(
            implementation_type,
            resource,
            params,
            function_spec,
        )

        # Pydantic's BaseModel is aware of all fields in both GeneratePrimitive and Primitive.
        # It properly processes the attributes to the correct classes and performs type-checking.
        super().__init__(
            type=PrimitiveType.GENERATE,
            implementation_type=implementation_type,
            dependencies=dependencies,
            context_keys=context_keys,
            inputs=inputs,  # type: ignore
            implementation_spec=spec,  # type: ignore
        )

        initialize_primitive_within_task(self)

    def _process_data(
        self, data: Union[CustomPrimitive, ContextKey]
    ) -> Tuple[List[str], List[ContextKey], List[Union[str, ContextKey]]]:
        """
        Verify and process the input data for the GeneratePrimitive instance.

        Parameters:
            data (Union[CustomPrimitive, ContextKey]): The input data to be verified and processed.

        Raises:
            InvalidInputException: If the input data is not of the expected type (CustomPrimitive or ContextKey).
        """
        from runllm.primitives.utils import process_primitive_inputs

        target_types = CustomPrimitive
        return process_primitive_inputs(data, target_types)

    def _construct_primitive_spec(
        self,
        implementation_type: ImplementationType,
        resource: Optional[Resource],
        params: Optional[GenerateParams],
        function_spec: Optional[FunctionSpec],
    ) -> Union[DefaultGenerateSpec, CustomSpec]:
        if implementation_type == ImplementationType.DEFAULT:
            if not (resource and params):
                raise Exception(
                    "Both resource and resource params need to be specified for the Default implementation"
                )
            self._verify_default_implementation(
                resource,
                params,
            )
            return DefaultGenerateSpec(
                resource_id=resource.id,
                params=params,
            )
        elif implementation_type == ImplementationType.CUSTOM:
            # We already verify the custom implementation in the decorator
            if not function_spec:
                raise InvalidConfigurationException(
                    "CUSTOM type implementation expects a function spec passed in."
                )

            return CustomSpec(
                function_spec=function_spec,
            )
        else:
            raise UnknownImplementationTypeException("Unknown type of implementation")

    def _verify_default_implementation(self, resource: Resource, params: GenerateParams):
        if resource.type == ResourceType.LLM:
            if not isinstance(params, GenerateParams):
                raise InvalidResourceParamsException(
                    "Parameters must be an instance of GenerateParams."
                )

            if resource.service == Service.OPENAI:
                if params.model not in OpenAIGPTModelType:
                    raise InvalidResourceParamsException(
                        f"Invalid model type {params.model} for OpenAI service. See runllm.constants.enums.OpenAIGPTModelType for supported model types."
                    )

        else:
            raise InvalidConfigurationException("Embed Primitive expects a LLM resource type.")
